#include <ros/ros.h>
#include <laserNode/Obstacles.h>
#include <GVG/GVGNode.h>
#include <visualization_msgs/Marker.h>
#include <visualization_msgs/MarkerArray.h>
#include <geometry_msgs/Point32.h>
#include <cstdlib>
#include <string>
#include "../GVGFollower.h"
#include "../GVGGraph.h"

GVGGraph  G; 
ros::Publisher obstacles_pub;
ros::Publisher obstacles_minima_pub;
ros::Publisher obstacle_names_pub;

void handle_obstacles(const laserNode::Obstacles::ConstPtr& msg);

int main(int argc, char **argv) {
  ros::init(argc, argv, "test_detect_meetpoints");
  ros::NodeHandle nh;
  ros::Subscriber obstacles_sub  = nh.subscribe("/indoor/laser_utils/closest_obstacles", 1, &handle_obstacles);
  obstacle_names_pub   = nh.advertise<visualization_msgs::MarkerArray>("obstacle_names", 1);
  obstacles_pub        = nh.advertise<visualization_msgs::MarkerArray>("obstacles", 1);
  obstacles_minima_pub = nh.advertise<visualization_msgs::MarkerArray>("obstacle_minima", 1);
  ros::spin();
  return 0;
}

void handle_obstacles(const laserNode::Obstacles::ConstPtr& msg) {
  std::vector<double>   possibleBearings;
  laserNode::Obstacles  meetpoint_obstacles;
  laserNode::Obstacles  obstacles = *msg;  
 
   
  bool onMeetPoint = GVGFollower::DetectMeetPoint(obstacles, possibleBearings, MEETPOINT_THRESHOLD, 
						  MEETPOINT_BEARING_ANGLE_DIFF, meetpoint_obstacles);
  if (onMeetPoint) {
    geometry_msgs::Point32 p; p.x = 0; p.y = 0;
    double closest_distance = obstacles.collection.at(0).min_distance;
    G.addMeetpoint(p, closest_distance, possibleBearings, meetpoint_obstacles);
    
    laserNode::Obstacles  obs_sorted_by_start = meetpoint_obstacles; 
    sort(obs_sorted_by_start.collection.begin(), obs_sorted_by_start.collection.end(), 
	 bind( &laserNode::Obstacle::start_index, _1 ) < bind( &laserNode::Obstacle::start_index, _2 ));
 
    std::string ns = "follow_edge_inputs";
    std::string frame_id = "/base_laser_link";
   
    visualization_msgs::MarkerArray ma;
    for (int i = 0; i < (int) obs_sorted_by_start.collection.size(); i++) {
      visualization_msgs::Marker m;
      m.header.frame_id = frame_id;
      m.header.stamp = ros::Time::now();
      m.ns = ns;
      m.id = i;
      m.type = visualization_msgs::Marker::LINE_STRIP;
      m.action = visualization_msgs::Marker::ADD;
      m.scale.x = 0.03;  
    
      if (i % 2 == 0) {
	m.color.a = 1.0;   m.color.r = 0.0;   m.color.g = 1.0;  m.color.b = 0.0;
      } else {
	m.color.a = 1.0;   m.color.r = 0.0;   m.color.g = 0.0;  m.color.b = 1.0;
      }

      std::vector<geometry_msgs::Point> pts;
      for (int j = 0; j < (int) obs_sorted_by_start.collection.at(i).surface.size(); j++) {
	geometry_msgs::Point pt;
	pt.x = obs_sorted_by_start.collection.at(i).surface.at(j).x;
	pt.y = obs_sorted_by_start.collection.at(i).surface.at(j).y;
	pts.push_back(pt);
      }
    
      m.points.assign(pts.begin(), pts.end());
      m.lifetime = ros::Duration(1.0/15.0);   //1/15th of a second
      ma.markers.push_back(m);
    }
  
    /*
     * Populating lines from the robot to the closest points of obstacles
     */
    visualization_msgs::MarkerArray ma_closest;
    for (int i = 0; i < (int) obs_sorted_by_start.collection.size(); i++) {
      visualization_msgs::Marker m;
      m.header.frame_id = frame_id;
      m.header.stamp = ros::Time::now();
      m.ns = ns;
      m.id = i + (int) obs_sorted_by_start.collection.size();
      m.type = visualization_msgs::Marker::LINE_LIST;
      m.action = visualization_msgs::Marker::ADD;
      m.scale.x = 0.01;  
    
      if (i % 2 == 0) {
	m.color.a = 1.0;   m.color.r = 0.0;   m.color.g = 1.0;  m.color.b = 0.0;
      } else {
	m.color.a = 1.0;   m.color.r = 0.0;   m.color.g = 0.0;  m.color.b = 1.0;
      }

      geometry_msgs::Point origin;
      origin.x = 0; origin.y = 0;
    
      geometry_msgs::Point closest_pt;
      closest_pt.x = obs_sorted_by_start.collection.at(i).closest_point.x;
      closest_pt.y = obs_sorted_by_start.collection.at(i).closest_point.y;
      m.points.push_back(origin);
      m.points.push_back(closest_pt);
      m.lifetime = ros::Duration(1.0/15.0);   //1/15th of a second
      ma_closest.markers.push_back(m);
    }

    /*
     * Populating lines from the robot to the possible directions it may follow
     */
    for (int i = 0; i < (int) obs_sorted_by_start.collection.size(); i++) {
      visualization_msgs::Marker m;
      m.header.frame_id = frame_id;
      m.header.stamp = ros::Time::now();
      m.ns = ns;
      m.id = i + (int) 2*obs_sorted_by_start.collection.size();
      m.type = visualization_msgs::Marker::TEXT_VIEW_FACING;
      m.action = visualization_msgs::Marker::ADD;
      m.color.a = 1.0;   m.color.r = 0.0;   m.color.g = 1.0;  m.color.b = 0.0;
      m.scale.z = 0.2;  
      if (i == 0) {
	double d0 = std::abs(-180 - obs_sorted_by_start.collection.at(i).min_bearing);
	double dend = std::abs(180 - obs_sorted_by_start.collection.back().min_bearing);
	m.text = boost::lexical_cast<std::string>((int)(d0 + dend));
      } else {
	m.text = boost::lexical_cast<std::string>((int)(std::abs(obs_sorted_by_start.collection.at(i).min_bearing - 
								 obs_sorted_by_start.collection.at(i-1).min_bearing)));
      }
      
      if (i == 0) {
	m.pose.position.x = cos(M_PI);
	m.pose.position.y = sin(M_PI);
      } else {
	m.pose.position.x = cos(possibleBearings.at(i-1)*M_PI/180.0);
	m.pose.position.y = sin(possibleBearings.at(i-1)*M_PI/180.0);
      }
 
      m.lifetime = ros::Duration(1.0/15.0);   //1/15th of a second
      ma_closest.markers.push_back(m);
    }

    /*
     * Populating angle differences between the different meetpoint obstacles
     */
    for (int i = 0; i < (int) possibleBearings.size(); i++) {
      visualization_msgs::Marker m;
      m.header.frame_id = frame_id;
      m.header.stamp = ros::Time::now();
      m.ns = ns;
      m.id = i + (int) 3*obs_sorted_by_start.collection.size();
      m.type = visualization_msgs::Marker::LINE_LIST;
      m.action = visualization_msgs::Marker::ADD;
      m.scale.x = 0.01;  
      m.color.a = 1.0;   m.color.r = 1.0;   m.color.g = 0.0;  m.color.b = 0.0;
      geometry_msgs::Point origin;
      origin.x = 0; origin.y = 0;
    
      geometry_msgs::Point possible_direction;
      possible_direction.x = 0.5*cos(possibleBearings.at(i)*M_PI/180.0);
      possible_direction.y = 0.5*sin(possibleBearings.at(i)*M_PI/180.0);
      
      m.points.push_back(origin);
      m.points.push_back(possible_direction);
      m.lifetime = ros::Duration(1.0/15.0);   //1/15th of a second
      ma_closest.markers.push_back(m);
    }

    /*
     * Populating names of obstacles
     */
    visualization_msgs::MarkerArray ma_names;
    for (int i = 0; i < (int) obs_sorted_by_start.collection.size(); i++) {
      visualization_msgs::Marker m;
      m.header.frame_id = frame_id;
      m.header.stamp = ma.markers.at(i).header.stamp;
      m.ns = ns;
      m.id = i + (int) 4*obs_sorted_by_start.collection.size();
      m.type = visualization_msgs::Marker::TEXT_VIEW_FACING;
      m.action = visualization_msgs::Marker::ADD;
      m.scale.z = 0.4;  
      m.text = boost::lexical_cast<std::string>(i);
      m.lifetime = ma.markers.at(i).lifetime;
      m.color = ma.markers.at(i).color;
      int size = (int) obs_sorted_by_start.collection.at(i).surface.size();
      m.pose.position.x = obs_sorted_by_start.collection.at(i).surface.at(floor(size/2.0)).x;
      m.pose.position.y = obs_sorted_by_start.collection.at(i).surface.at(floor(size/2.0)).y;
      m.pose.position.z = obs_sorted_by_start.collection.at(i).surface.at(floor(size/2.0)).z;
      ma_names.markers.push_back(m);
    }
  
    obstacles_pub.publish(ma);
    obstacles_minima_pub.publish(ma_closest);
    obstacle_names_pub.publish(ma_names);
  }
}
