#include <geometry_msgs/Point32.h>
#include <Eigen/Core>
#include <random>
#include <string>
#include <cmath>
#include <gtest/gtest.h>
#include "../GVGGraph.h"

TEST(TestSuite, test_pi_over_four){
  std::vector<geometry_msgs::Point32> e1;
  std::vector<geometry_msgs::Point32> e2;
  int N = 100;

  std::default_random_engine generator;
  std::normal_distribution<double> randn(0, 1);

  for (int i = 0; i < N; i++) {
    geometry_msgs::Point32 p1;
    p1.x = i/((double) N) * 5;
    p1.y = p1.x * p1.x + 0.2*randn(generator);
    e1.push_back(p1);

    geometry_msgs::Point32 p2;
    p2.x = i/((double) N) * 5;
    p2.y = p2.x * p2.x + 0.4*randn(generator);
    e2.push_back(p2);
  }

  double theta = M_PI/4;
  Eigen::Rotation2Dd R(theta);
  Eigen::Vector2d    t; t << -1, 1;

  for (int i = 0; i < N; i++) {
    Eigen::Vector2d v;  v << e2.at(i).x, e2.at(i).y;
    v = R*v + t;
    e2.at(i).x = v.x();
    e2.at(i).y = v.y();
  }

  double tx, ty;
  double error = estimate_transformation(e1, e2, theta, tx, ty);
 
  ROS_INFO("theta %f (should be pi/4)", theta);
  ROS_INFO("tx, ty = %f, %f (should be -1, 1)", tx, ty);
  EXPECT_TRUE(std::abs(error) < 50);
  EXPECT_TRUE(std::abs(theta - M_PI/4.0) < 0.2);
  EXPECT_TRUE(std::abs(tx + 1) < 0.2);
  EXPECT_TRUE(std::abs(ty - 1) < 0.2);
}


TEST(TestSuite, test_add_meetpoint){
  GVGGraph G;
  
  geometry_msgs::Point32 p;
  laserNode::Obstacles obs;
  laserNode::Obstacle o;
  obs.collection.push_back(o);
  obs.collection.push_back(o);
  obs.collection.push_back(o);

  std::vector<double> possible_bearings;
  possible_bearings.push_back(100);
  possible_bearings.push_back(-90);
  
  GVG::GVGNode v = G.addMeetpoint(p, 0.0, possible_bearings, obs);
  
  EXPECT_TRUE(v.node_id == 0);
  EXPECT_TRUE(v.degree == 3);
  EXPECT_TRUE((int) v.edge_angle_diffs.size() == 3);
  EXPECT_TRUE(std::abs(v.edge_angle_diffs[0] - 90) < 1);
  EXPECT_TRUE(std::abs(v.edge_angle_diffs[1] - 190) < 1);
  EXPECT_TRUE(std::abs(v.edge_angle_diffs[2] - 80) < 1);
}

// Run all the tests that were declared with TEST()
int main(int argc, char **argv){
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
