#ifndef GVG_GRAPH_H
#define GVG_GRAPH_H

#include <ros/ros.h>
#include <string>
#include <boost/config.hpp>
#include <vector>
#include <boost/graph/dijkstra_shortest_paths.hpp>
#include <boost/utility.hpp>                // for boost::tie
#include <boost/graph/adjacency_list.hpp>
#include <geometry_msgs/Point32.h>
#include <laserNode/Obstacles.h>
#include "GVGMapper/GVGNode.h"
#include "GVGMapper/GVGEdgeMsg.h"
#include <Eigen/Core>
#include <Eigen/Geometry>
#include <unsupported/Eigen/NonLinearOptimization>
#include <BasicGeometry.h>
#include "GVGMapper/AddMeetpoint.h"
#include "GVGMapper/AddEndpoint.h"
#include "GVGMapper/ExtendEdge.h"
#include "GVGMapper/RetrieveBearings.h"
#include "GVGMapper/RetrievePath.h"

/*
 * Radius of search for an existing vertex when we are trying to perform loop closure. In m.
 */
#define SAME_VERTEX_RADIUS 0.5 

class EdgeBearing;

class UncertaintyEllipse {
 public:
  
  geometry_msgs::Point32 center;
  double halfaxis1;
  double halfaxis2;
  double angle; // Radians

  bool contains(geometry_msgs::Point32& p);

};

class GVGVertex {
 public:

  int node_id;
 
  geometry_msgs::Point32 p;

  // For meetpoints, the number of obstacles that 
  // defined it. For endpoints, 1.   
  int expected_degree;

  // Proximity to the closest obstacle
  double closest_distance; 

  // Angle differences between edges adjacent to this vertex (in rads)
  //std::vector<double> edge_angle_diffs; 
  double vertex_angle;

  laserNode::Obstacles surrounding_obstacles;

  // List of possible bearings (absolute angle in rads) from vertex
  std::vector<EdgeBearing> possible_bearings;

  // True iff these two vertices are estimated to be similar 
  bool similar_vertex(GVGVertex& v);

  // edge angle diffs (Use the angle edge diff stored in bearings, relative to edge 0)
  std::vector<double> edge_angle_diffs;

};

class GVGEdge {
 public:

  int source;
  int target;

  int edge_id;
  std::vector<geometry_msgs::Point32> line;

  double length;

  // True iff these two edges are estimated to be similar
  bool similar_edge(GVGEdge& e);
};


typedef boost::property<boost::edge_weight_t, double> EdgeWeightProperty;

typedef boost::adjacency_list<boost::vecS, boost::vecS, 
                              boost::undirectedS, GVGVertex, EdgeWeightProperty> Graph;

typedef boost::graph_traits<Graph>::vertex_descriptor     Vertex;
typedef boost::graph_traits<Graph>::edge_descriptor       Edge;

typedef boost::graph_traits<Graph>::vertex_iterator       Vertex_iter;
typedef boost::graph_traits<Graph>::edge_iterator         Edge_iter;
typedef boost::graph_traits<Graph>::adjacency_iterator    Adjacency_iter;
typedef boost::graph_traits<Graph>::out_edge_iterator     Out_edge_iter;

class EdgeBearing {
 public:
  int bearing_id;
  int edge_id;
  double edge_angle_diff;
  bool visited;

};

class GVGGraph {
 public:

  GVGGraph(void);

  /*
   * edge_angle_diffs1 should be the current vertex we are trying to match, and edge_angle_diffs2 should iterate over all the vertices in our current GVG Graph.
   * Returns bearing ID which minimizes the error that matches the two vertices. -1 if the # of bearings differ (not same vertex).
   */
  std::vector<double> edge_angle_diffs_error(std::vector<double>& edge_angle_diffs1, std::vector<double>& edge_angle_diffs2);

  /*
   * Adds a new meetpoint to the GVG graph. p is the location of the meetpoint in world  
   * coordinates, closest_distance is the distance to the (at least) three closest obstacles
   * that define the meetpoint. possible_bearings are the angles that the robot can turn by
   * to choose another GVG edge. These angles are given in laser coordinates, i.e. with respect 
   * to the x-axis, positive angles on its right, negative on its left. The angles are in rads.   
   */
  bool addMeetpoint(GVGMapper::AddMeetpoint::Request& req, GVGMapper::AddMeetpoint::Response& res);
  
  /*
   * Adds a new endpoint to the GVG graph. p is the location of the endpoint in world  
   * coordinates, closest_distance is the distance to the closest obstacle that defines the 
   * endpoint.    
   */
  bool addEndpoint(GVGMapper::AddEndpoint::Request& req, GVGMapper::AddEndpoint::Response& res);

  /*
   * The robot is currently traversing an edge, so just extend that edge in the 
   * internal representation. 
   */
  bool extendGVGEdge(GVGMapper::ExtendEdge::Request& req, GVGMapper::ExtendEdge::Response& res);
  
  /*
   * Service call for Planner, returns the list of possible bearings at vertex node_id.
   */
  bool retrieveBearings(GVGMapper::RetrieveBearings::Request& req, GVGMapper::RetrieveBearings::Response& res);

  bool retrievePath(GVGMapper::RetrievePath::Request& req, GVGMapper::RetrievePath::Response& res);

 private:
  Graph G;
  Out_edge_iter  out,    out_end;
  Adjacency_iter nhd,    nhd_end;
  int next_node_id;
  int next_edge_id;
  std::vector<int> unexploredEdgesCount;

  std_msgs::Header header;

  // Physical distance between laser range finder and robot center
  double laser_offset_x;
  double laser_offset_y;
  
  // Bearing we arrive at meetpoint from
  int arriving_bearing;

  // ROS launch params stored
  double closest_distance_threshold;
  double edge_length_threshold;
  int exploration_policy;
  int vertex_matching_policy;

  GVGEdge currentEdge;
  Vertex lastVertex;
  // List of all edges in the graph
  std::vector<GVGEdge> edges_list;
  
  Vertex nullVertex();
  Vertex addVertex(GVGVertex& v, double robot_angle, geometry_msgs::PoseWithCovariance& pose);
  std::string vertex_state(int node_id);
  std::vector<int> computeDijkstra(Vertex v);
  
  // ID of the closest vertex with unexplored edges when path planning.
  int closest_unexplored_vertex;
  // length of last edge traversed, for vertex matching and relocalization purposes
  double last_edge_length;
  
  bool relocalize;
  double robot_angle_correction;
  std::vector<double> local_possible_bearings;
  
  // Ellipse information
  UncertaintyEllipse ellipse;
 
  /* The (at least three) objects that are going to define a meetpoint need to have
     closest point ranges that differ no more than this quantity. In m. */
  double MEETPOINT_THRESHOLD;

  /* Services provided by GVGMapper pkg to update the graph */
  ros::NodeHandle nh;

  ros::ServiceServer add_meetpoint_srv;
  ros::ServiceServer add_endpoint_srv;
  ros::ServiceServer extend_edge_srv;
  ros::ServiceServer retrieve_bearings_srv;
  ros::ServiceServer retrieve_path_srv;
  ros::ServiceClient process_meetpoint_cln;
  ros::Publisher     gvg_node_pub;
  ros::Publisher     gvg_edge_pub;
};
  
double abs_angle(double theta, double phi);

double absoluteToRelativeAngle(double robot_angle, double absolute_angle);

// Generic functor
template<typename _Scalar, int NX=Eigen::Dynamic, int NY=Eigen::Dynamic>
struct Functor {

  typedef _Scalar Scalar;
  enum {
    InputsAtCompileTime = NX,
    ValuesAtCompileTime = NY
  };
  typedef Eigen::Matrix<Scalar,InputsAtCompileTime,1> InputType;
  typedef Eigen::Matrix<Scalar,ValuesAtCompileTime,1> ValueType;
  typedef Eigen::Matrix<Scalar,ValuesAtCompileTime,InputsAtCompileTime> JacobianType;

  const int m_inputs, m_values;

  Functor() : m_inputs(InputsAtCompileTime), m_values(ValuesAtCompileTime) {}
  Functor(int inputs, int values) : m_inputs(inputs), m_values(values) {}

  int inputs() const { return m_inputs; }
  int values() const { return m_values; }

  // you should define that in the subclass :
  // void operator() (const InputType& x, ValueType* v, JacobianType* _j=0) const;
};

struct lm_functor_without_jacobian : Functor<double> {
  
 std::vector<geometry_msgs::Point32> *e1;
 std::vector<geometry_msgs::Point32> *e2; 

 lm_functor_without_jacobian(std::vector<geometry_msgs::Point32>& e1_, 
			     std::vector<geometry_msgs::Point32>& e2_): 

      // 3 is the numbers of unknowns in the problem (theta, tx, ty) and e1.size == e2.size
      // is the number of given observation points.   
      Functor<double>(3, e1_.size()) {

    this->e1 = &e1_;
    this->e2 = &e2_;
  }

  int operator() (const Eigen::VectorXd &vars, Eigen::VectorXd &fvec) const {
    
    int N = e1->size();
    assert(values() == N);
    
    double theta = vars[0];
    double tx    = vars[1];
    double ty    = vars[2];
    
    for (int i = 0; i < N; i++) {
      Eigen::Vector2d v2; v2 << e2->at(i).x, e2->at(i).y;
      Eigen::Vector2d v1; v1 << e1->at(i).x, e1->at(i).y;
      Eigen::Vector2d t;   t << tx, ty;
      Eigen::Rotation2Dd R(theta);
      
      Eigen::Vector2d error = v2 - (R*v1 + t); 
      
      fvec(i) = error.norm();
    }

    return 0;
  }
};

#endif 
