#include <GVGPlanner/SelectBearing.h>
#include "Planner.h"
#include <GVGMapper/RetrieveBearings.h>
#include <GVGMapper/RetrievePath.h>
#include <localizer/MinUncertainty.h>
#include <localizer/MaxUncertainty.h>

/*
 * Pass meetpoint/endpoint signatures from Follower to Mapper.
 */

Planner::Planner() {
  this->state = 0;
  this->select_bearing_srv = nh.advertiseService("select_bearing", &Planner::SelectBearingGVG, this);
  this->retrieve_bearings_cln = nh.serviceClient<GVGMapper::RetrieveBearings>("/retrieve_bearings");
  this->retrieve_path_cln = nh.serviceClient<GVGMapper::RetrievePath>("/retrieve_path");
  this->min_uncertainty_cln = nh.serviceClient<localizer::MinUncertainty>("/indoor/gvg/MinUncertainty");
}

/*
 * Query the Mapper for information about current node and retrieve selected bearing for next exploration.
 * If the meetpoint has all visited edges, retrieve path plan from Mapper to return to previous meetpoint with unexplored edges.
 */
bool Planner::SelectBearingGVG(GVGPlanner::SelectBearing::Request& req, GVGPlanner::SelectBearing::Response& res) {
  // Exploration state
  if (state == 0) {
    /* Retrieve available bearings from GVGMapper */
    GVGMapper::RetrieveBearings srv;
    srv.request.node_id = req.node_id;
    srv.request.target = -1;
    if (!retrieve_bearings_cln.call(srv)) {
      ROS_WARN("Could not retrieve bearings from Mapper!");
      res.success = false;
      return true;
    }
    
    // Relocalize, pathplan to node with MIN uncertainty
    if (srv.response.relocalize == 1) {      
      ROS_INFO("Robot uncertainty too large. Path planning to minimum uncertainty vertex.");
      localizer::MinUncertainty minsrv;
      if (!min_uncertainty_cln.call(minsrv)) {
        ROS_WARN("Could not get minimum uncertainty vertex from localizer!");
        res.success = false;
        return true;
      }
      GVGMapper::RetrievePath pathsrv;
      pathsrv.request.source = req.node_id;
      pathsrv.request.target = minsrv.response.id;
      if (!retrieve_path_cln.call(pathsrv)) {
        ROS_WARN("Could not retrieve path plan from Mapper!");
        res.success = false;
        return true;
      }
      
      vertex_list = pathsrv.response.vertex_list;
      this->state = 1;
      
      target = vertex_list.back();
      vertex_list.pop_back();

      // Retrieve the bearing we need to head next.
      GVGMapper::RetrieveBearings srv;
      srv.request.node_id = req.node_id;
      srv.request.target = target;
      if (!retrieve_bearings_cln.call(srv)) {
        ROS_WARN("Could not retrieve bearings from Mapper!");
        res.success = false;
        return true;
      }
      double chosen_bearing = srv.response.chosen_bearing;

      ROS_INFO("Requested direction from GVG Follower: %.2f", chosen_bearing * 180.0/M_PI);
      res.selected_bearing = chosen_bearing;

      res.success = true; 
      if (vertex_list.size() == 0) this->state = 0;
      return true;
    }
    
    // Random relocalization, path plan to MAX uncertainty vertex
    if (srv.response.relocalize == 2) {      
      ROS_INFO("Random relocalization. Path planning to maximum uncertainty vertex.");
      localizer::MaxUncertainty maxsrv;
      if (!max_uncertainty_cln.call(maxsrv)) {
        ROS_WARN("Could not get maximum uncertainty vertex from localizer!");
        res.success = false;
        return true;
      }
      GVGMapper::RetrievePath pathsrv;
      pathsrv.request.source = req.node_id;
      pathsrv.request.target = maxsrv.response.id;
      if (!retrieve_path_cln.call(pathsrv)) {
        ROS_WARN("Could not retrieve path plan from Mapper!");
        res.success = false;
        return true;
      }
      
      vertex_list = pathsrv.response.vertex_list;
      this->state = 1;
      
      target = vertex_list.back();
      vertex_list.pop_back();

      // Retrieve the bearing we need to head next.
      GVGMapper::RetrieveBearings srv;
      srv.request.node_id = req.node_id;
      srv.request.target = target;
      if (!retrieve_bearings_cln.call(srv)) {
        ROS_WARN("Could not retrieve bearings from Mapper!");
        res.success = false;
        return true;
      }
      double chosen_bearing = srv.response.chosen_bearing;

      ROS_INFO("Requested direction from GVG Follower: %.2f", chosen_bearing * 180.0/M_PI);
      res.selected_bearing = chosen_bearing;

      res.success = true; 
      if (vertex_list.size() == 0) this->state = 0;
      return true;
    }

    double chosen_bearing = srv.response.chosen_bearing;

    // Still unexplored edges to visit
    if (chosen_bearing != NULL) {
      ROS_INFO("Requested direction from GVG Follower: %.2f", chosen_bearing * 180.0/M_PI);
      res.selected_bearing = chosen_bearing;

      res.success = true; 
      return true;
    }
    // Explored all edges at this vertex, path plan to previously nonexplored edges vertex.
    else {
      ROS_INFO("All edges explored at this vertex. Requesting path to previously non-explored edges vertex.");
      GVGMapper::RetrievePath pathsrv;
      pathsrv.request.source = req.node_id;
      pathsrv.request.target = -1;
      if (!retrieve_path_cln.call(pathsrv)) {
        ROS_WARN("Could not retrieve path plan from Mapper!");
        res.success = false;
        return true;
      }

      // Stop the robot
      if ((int) pathsrv.response.vertex_list.size() == 0) {
        ROS_WARN("Entire map has been explored.");
        nh.setParam("/indoor/gvg/agent/do_move", false);
        res.success = true;
        return true;
      }

      vertex_list = pathsrv.response.vertex_list;
      this->state = 1;
      
      target = vertex_list.back();
      vertex_list.pop_back();

      // Retrieve the bearing we need to head next.
      GVGMapper::RetrieveBearings srv;
      srv.request.node_id = req.node_id;
      srv.request.target = target;
      if (!retrieve_bearings_cln.call(srv)) {
        ROS_WARN("Could not retrieve bearings from Mapper!");
        res.success = false;
        return true;
      }
      double chosen_bearing = srv.response.chosen_bearing;

      ROS_INFO("Requested direction from GVG Follower: %.2f", chosen_bearing * 180.0/M_PI);
      res.selected_bearing = chosen_bearing;

      res.success = true; 
      if (vertex_list.size() == 0) this->state = 0;
      return true;
    }
  
  // Pathplanning state
  } else {
    if (target != req.node_id) {
      ROS_WARN("Robot did not reach the planned target. Restarting path planning.");
      GVGMapper::RetrievePath pathsrv;
      pathsrv.request.source = req.node_id;
      if (!retrieve_path_cln.call(pathsrv)) {
        ROS_WARN("Could not retrieve path plan from Mapper!");
        res.success = false;
        return true;
      }
      vertex_list = pathsrv.response.vertex_list;
      this->state = 1;
      
      target = vertex_list.back();
      vertex_list.pop_back();

      // Retrieve the bearing we need to head next.
      GVGMapper::RetrieveBearings srv;
      srv.request.node_id = req.node_id;
      srv.request.target = target;
      if (!retrieve_bearings_cln.call(srv)) {
        ROS_WARN("Could not retrieve bearings from Mapper!");
        res.success = false;
        return true;
      }
      double chosen_bearing = srv.response.chosen_bearing;

      ROS_INFO("Requested direction from GVG Follower: %.2f", chosen_bearing * 180.0/M_PI);
      res.selected_bearing = chosen_bearing;

      res.success = true; 
      if (vertex_list.size() == 0) this->state = 0;
      return true;
    }
    target = vertex_list.back();
    vertex_list.pop_back();

    // Retrieve the bearing we need to head next.
    GVGMapper::RetrieveBearings srv;
    srv.request.node_id = req.node_id;
    srv.request.target = target;
    if (!retrieve_bearings_cln.call(srv)) {
      ROS_WARN("Could not retrieve bearings from Mapper!");
      res.success = false;
      return true;
    }
    double chosen_bearing = srv.response.chosen_bearing;

    ROS_INFO("Requested direction from GVG Follower: %.2f", chosen_bearing * 180.0/M_PI);
    res.selected_bearing = chosen_bearing;

    res.success = true; 
    if (vertex_list.size() == 0) this->state = 0;
    return true;  
  }
}

double Planner::absoluteToRelativeAngle(double robot_angle, double absolute_angle) {
  double relative_angle = absolute_angle;
  if (robot_angle > 0) {
    if (relative_angle < 0) relative_angle = -relative_angle;
    else {
      if (relative_angle > M_PI) relative_angle = 2*M_PI - relative_angle;
      else relative_angle = -relative_angle;
    }
  } else {
    if (relative_angle > 0) relative_angle = -relative_angle;
    else {
      if (relative_angle < -M_PI) relative_angle = -2*M_PI - relative_angle;
      else relative_angle = -relative_angle;
    }
  }
  return relative_angle;
}
