#include <iostream>
#include <stdio.h>
#include <limits>
#include <cassert>
#include <cmath>
#include <algorithm>
#include "BasicGeometry.h"
#include "LaserUtils.h"

using namespace std;

LaserUtils::LaserUtils(std::string& robot_name) {
  laserSub   = nh.subscribe("/indoor/base_scan", 1, &LaserUtils::laserScanFiltering, this);
  filteredLaserSub = nh.subscribe("/indoor/laser_utils/filtered_scan", 1, &LaserUtils::handleLaserScan, this);
  allObsPub  = nh.advertise<laserNode::Obstacles>("/indoor/laser_utils/closest_obstacles", 1);
  filteredScanPub = nh.advertise<sensor_msgs::LaserScan>("/indoor/laser_utils/filtered_scan", 1);
}

double LaserUtils::getRange(const sensor_msgs::LaserScan::ConstPtr& msg, int i) { 
  float r = msg->ranges.at(i);
  assert(i >= 0 && i <= countScans - 1);
  if (r > 5.0) return msg->range_max;
  return r;
}

double LaserUtils::getBearing(const sensor_msgs::LaserScan::ConstPtr& msg, int i) { 
  double theta = i*msg->angle_increment + msg->angle_min;
  assert(msg->angle_max > msg->angle_min);
  assert(msg->angle_max > M_PI/2.0 && msg->angle_min < -M_PI/2.0);
  assert(i >= 0 && i <= countScans - 1);
  assert(theta >= msg->angle_min && theta <= msg->angle_max);
  return theta;
}

geometry_msgs::Point32 LaserUtils::toPoint32(const sensor_msgs::LaserScan::ConstPtr& msg, int i) {
  geometry_msgs::Point32 p;
  double r = getRange(msg, i);
  double theta = getBearing(msg, i);
  p.x = r * cos(theta);
  p.y = r * sin(theta);
  return p;
}

void LaserUtils::laserScanFiltering(const sensor_msgs::LaserScan::ConstPtr& msg) {
  sensor_msgs::LaserScan scan;
  scan.header = msg->header;
  scan.angle_min = msg->angle_min;
  scan.angle_max = msg->angle_max;
  scan.angle_increment = msg->angle_increment;
  scan.time_increment = msg->time_increment;
  scan.scan_time = msg->scan_time;
  scan.range_min = msg->range_min;
  scan.range_max = msg->range_max;
  for (int i = 0; i < (int) msg->ranges.size(); i++) {
    scan.ranges.push_back(msg->ranges.at(i));
    if (scan.ranges.at(i) < scan.range_min) {
      scan.ranges.at(i) = scan.range_max;
    }
  }
  filteredScanPub.publish(scan);
}

void LaserUtils::handleLaserScan(const sensor_msgs::LaserScan::ConstPtr& msg) {
  countScans = std::floor((msg->angle_max - msg->angle_min)/msg->angle_increment);

  laserNode::Obstacles obs_msg;
  
  bool found_good_obstacle_set = false;
  double epsilon = 0.5;  //in meters
  double minRange = 0;

  while (minRange + epsilon < msg->range_max && !found_good_obstacle_set) {
    obs_msg.collection.clear();
    getClosestObstaclesViaAnnulus(obs_msg, epsilon, msg);
    
    if (obs_msg.collection.empty()) {
      epsilon += 0.5;
   
    } else if (obs_msg.collection.size() == 1){ 
      minRange = obs_msg.collection.at(0).min_distance;
      epsilon += 0.5;
    
    } else if (obs_msg.collection.size() == 3) {
      laserNode::Obstacles obs_msg2;
      getClosestObstaclesViaAnnulus(obs_msg2, epsilon*1.5, msg);
      if (obs_msg2.collection.size() > 3) {
        obs_msg = obs_msg2;
      }
      
      found_good_obstacle_set = true;

    } else {
      found_good_obstacle_set = true;
    }
  }
  
  std::vector<int> obstacles_to_merge;
  std::vector<int> obstacles_indices;
  // Sort all the obstacles' indices
  for (int i = 0; i < (int) obs_msg.collection.size(); i++) {
    obstacles_indices.push_back(obs_msg.collection.at(i).start_index);
    obstacles_indices.push_back(obs_msg.collection.at(i).end_index);
  }
  std::sort(obstacles_indices.begin(), obstacles_indices.end());
  // Check that all the possible bearings are indeed accessible
  for (int j = 1; j < (int) obstacles_indices.size() - 1; j += 2) {
    int current_index = obstacles_indices.at(j);
    int end_index = obstacles_indices.at(j+1);
    int next = current_index + 1;
    bool opening = false;
    while (current_index < end_index) {
      while ((getRange(msg, next) >= msg->range_max)||(getRange(msg, next) < msg->range_min)) next++;
      geometry_msgs::Point32 start = toPoint32(msg, current_index);
      geometry_msgs::Point32 end = toPoint32(msg, next);
      current_index = next;
      next = current_index + 1;
      if (norm(start, end) > ROBOT_DIAM) {
        opening = true;
        break;
      }
    }
    // No opening available, keep track of which obstacles need to be merged
    if (!opening) {
      int first, second;
      for (int k = 0; k < (int) obs_msg.collection.size(); k++) {
        if (obstacles_indices.at(j) == obs_msg.collection.at(k).end_index) first = k;
        else if (obstacles_indices.at(j+1) == obs_msg.collection.at(k).start_index) second = k;
      }
      obstacles_to_merge.push_back(first);
      obstacles_to_merge.push_back(second);
    }
  }

  if (!obstacles_to_merge.empty()) {

    std::vector<laserNode::Obstacle> collection;
    // Merge the obstacles by fixing the obstacle collection list
    for (int i = 0; i < (int) obstacles_to_merge.size(); i += 2) {
      laserNode::Obstacle first;
      if ((i > 0) && (obstacles_to_merge.at(i) == obstacles_to_merge.at(i-1))) {
        first = collection.back();
        collection.pop_back();
      }
      else first = obs_msg.collection.at(obstacles_to_merge.at(i));
      laserNode::Obstacle second = obs_msg.collection.at(obstacles_to_merge.at(i+1));
      laserNode::Obstacle result;
      result.start_index = first.start_index;
      result.end_index = second.end_index;
      if (first.min_distance < second.min_distance) {
        result.min_distance = first.min_distance;
        result.min_bearing = first.min_bearing;
        result.min_index = first.min_index;
        result.closest_point = first.closest_point;
      } else {
        result.min_distance = second.min_distance;
        result.min_bearing = second.min_bearing;
        result.min_index = second.min_index;
        result.closest_point = second.closest_point;
      }
      for (int j = 0; j < (int) first.surface.size(); j++) result.surface.push_back(first.surface.at(j));
      int index = first.end_index + 1;
      while (index < second.start_index) {
        geometry_msgs::Point32 p;
        p = toPoint32(msg, index);
        result.surface.push_back(p);
        index++;
      }
      for (int j = 0; j < (int) second.surface.size(); j++) result.surface.push_back(second.surface.at(j));
      collection.push_back(result);
    }

    // Add in all the unmodified obstacles and sort
    for (int i = 0; i < (int) obs_msg.collection.size(); i++) {
      bool unmodified = true;
      for (int j = 0; j < (int) obstacles_to_merge.size(); j++) {
        if (obstacles_to_merge.at(j) == i) {
          unmodified = false;
          break;
        }
      }
      if (unmodified) collection.push_back(obs_msg.collection.at(i));
    }

    // Fix the obstacles for sharp corners and overwrite the obstacles collection (to avoid detecting non-existent meetpoint)
    if (obs_msg.collection.size() - (obstacles_to_merge.size()/2) > 1) {
      obs_msg.collection.clear();
      obs_msg.collection.assign(collection.begin(), collection.end());
      std::sort(obs_msg.collection.begin(), obs_msg.collection.end(), bind( &laserNode::Obstacle::start_index, _1 ) < bind( &laserNode::Obstacle::start_index, _2 ));  
    }
    // Otherwise, we have a case where we have a very distant endpoint (so we want to follow GVG until we find the end of the endpoint) -- we store the merged obstacles
    else {
      obs_msg.merged_collection.clear();
      obs_msg.merged_collection.assign(collection.begin(), collection.end());
      std::sort(obs_msg.merged_collection.begin(), obs_msg.merged_collection.end(), bind( &laserNode::Obstacle::start_index, _1 ) < bind( &laserNode::Obstacle::start_index, _2 ));  
    }
  }

  if (obs_msg.collection.size() == 1) {
    obs_msg.merged_collection.assign(obs_msg.collection.begin(), obs_msg.collection.end());
  }

  obs_msg.countScans = countScans;
  obs_msg.seq = msg->header.seq;
  obs_msg.stamp = msg->header.stamp;
  obs_msg.angle_min = msg->angle_min;
  obs_msg.angle_max = msg->angle_max;
  obs_msg.angle_increment = msg->angle_increment;
  allObsPub.publish(obs_msg);
  ros::spinOnce();
}


/* Clusters laser scans into obstacles using an annulus around the closest obstacle. */
void LaserUtils::getClosestObstaclesViaAnnulus(laserNode::Obstacles& annulus_obstacles, double epsilon, 
					       const sensor_msgs::LaserScan::ConstPtr& msg) {
  
  double minRange = msg->range_max + 1;
  for (int i = 0; i < countScans; i++) {
    if (getRange(msg, i) < minRange) {
      minRange = getRange(msg, i);
    }
  }

  int end_index = 0;
  while (end_index < countScans) {
    
    if (end_index < countScans && getRange(msg, end_index) < minRange + epsilon) {  // new obstacle starts
     
      laserNode::Obstacle obs;
      obs.start_index  = end_index;
      obs.min_distance = msg->range_max + 1;
      obs.min_bearing  = msg->angle_min - 1;
      obs.min_index    = -1;
      obs.end_index    = obs.start_index;
          
      do {
	      double bearing = getBearing(msg, end_index);
	      double range   = getRange(msg, end_index);
	      geometry_msgs::Point32 p = toPoint32(msg, end_index);
	      obs.surface.push_back(p);
	      obs.end_index = end_index;

	      if (range < obs.min_distance) {
	        obs.min_distance = range;
	        obs.min_bearing = bearing;
	        obs.min_index = end_index;
	        obs.closest_point = p;
	      }
            	
	      end_index++;
      } while (end_index < countScans && getRange(msg, end_index) < minRange + epsilon);

      assert(obs.start_index <= obs.end_index);
      
      if ((int) obs.surface.size() >= 2 && annulus_obstacles.collection.empty()) {
	      annulus_obstacles.collection.push_back(obs);

      } else if ((int) obs.surface.size() >= 2) {

	    if (norm(annulus_obstacles.collection.back().surface.back(), obs.surface.front()) < ROBOT_DIAM) {
	      // robot can't go through, consider them one continuous obstacle
	      assert (annulus_obstacles.collection.back().end_index < obs.start_index);
	      if (annulus_obstacles.collection.back().min_distance > obs.min_distance) {
	        annulus_obstacles.collection.back().min_distance = obs.min_distance;
	        annulus_obstacles.collection.back().min_bearing = obs.min_bearing;
	        annulus_obstacles.collection.back().closest_point = obs.closest_point;
	        annulus_obstacles.collection.back().min_index = obs.min_index;
	      }
	      
	      for (int i = annulus_obstacles.collection.back().end_index + 1; i <= obs.end_index; i++) {
	        annulus_obstacles.collection.back().end_index = i;
	        annulus_obstacles.collection.back().surface.push_back(toPoint32(msg, i));
	      }
	
        } else {
	        annulus_obstacles.collection.push_back(obs);
	      }
      }

    } else {
      end_index++;
    }
  }

  std::sort(annulus_obstacles.collection.begin(), annulus_obstacles.collection.end(), 
	    bind( &laserNode::Obstacle::start_index, _1 ) < bind( &laserNode::Obstacle::start_index, _2 ));

}
