#include <iostream>
#include <limits>
#include "localizerEKF_2D.h"
#include <tf/transform_listener.h>


localizerEKF_2D::localizerEKF_2D(std::string& robot_name):
  X(3),P(3,3) {
  ROS_INFO("localizerEKF_2D::localizerEKF_2D ");

  std::string output_frame;
  std::string theMap_frame;
  
  nh.getParam("/indoor/gvg/simulation", _Simulation);
  
  if (_Simulation)
    odomSub = nh.subscribe("/indoor/odom",1,&localizerEKF_2D::handleOdomSim, this);
  else 
    odomSub = nh.subscribe("/odom_combined",1,&localizerEKF_2D::handleOdomTurtle, this);
  startService =  nh.advertiseService("StartFilter"    , &localizerEKF_2D::processStart,this);
  updateService = nh.advertiseService("UpdateFilter"  , &localizerEKF_2D::processMeetpoint,this);
  maxUncService = nh.advertiseService("MaxUncertainty", &localizerEKF_2D::processMaxUnc,this);
  minUncService = nh.advertiseService("MinUncertainty", &localizerEKF_2D::processMinUnc,this);

  nh.param("output_frame", output_frame, std::string("odom_combined"));
  posePub = nh.advertise<nav_msgs::Odometry>(output_frame, 10);

  nh.param("theMap_frame", theMap_frame, std::string("theMap"));

  mapPub = nh.advertise<localizer::GVGmap>(theMap_frame, 10);
  
  filterOn=true;
  initOdom=false;
  nL=0; // number of landmarks
  // EKF constants
  X=VectorXd(3); // The state vector initialized to the pose of the robot.
  X<< 0,0,0;

  P=MatrixXd::Zero(3,3); // The covariance matrix

  // The model noise covariance (linear and angular velocity)
  Qr<< _Wvv, _Wvw,
    _Wvw, _Www;
  CF<<1,0,0,
    0,1,0,
    0,0,1;
  J<<0, -1,
    1, 0;
  //cout<<"J"<<J<<endl;
  R<< 0.02527455743763,   0.000014869773007,
    0.000014869773007,   0.02476827355463;
}

void localizerEKF_2D::handleOdomTurtle(const geometry_msgs::PoseWithCovarianceStamped::ConstPtr& odom) {

  double x=odom->pose.pose.position.x;
  double y=odom->pose.pose.position.y;
  //  double yaw=acos(odom->pose.pose.orientation.w)*2.0;
  tf::Pose pose;
  tf::poseMsgToTF(odom->pose.pose, pose);
  double yaw=tf::getYaw(pose.getRotation());
  // ROS_INFO("localizerEKF_2D::handleOdom received [%lf,%lf,%lf]",x,y,yaw*180/M_PI);
  if(!initOdom) {
    oldX=x;
    oldY=y;
    oldYaw=yaw;
    initOdom=true;
    oldStamp=odom->header.stamp;
    X<<x,y,yaw;
    //  ROS_INFO("localizerEKF_2D::handleOdom X initialized to [%lf,%lf,%lf]",X(0),X(1),X(2)*180/M_PI);
    return;
  }
  if(filterOn) {
    if((oldX==x)&&(oldY==y)&&(oldYaw==yaw)){
      oldStamp=odom->header.stamp;
      //      ROS_INFO("localizerEKF_2D::handleOdom same position ending");
      return;
    }
    // Calculate the measurement
    double dx=x-oldX;
    double dy=y-oldY;
    double dYaw=angleDiff(yaw,oldYaw);
    double dt=odom->header.stamp.toSec()-oldStamp.toSec();
    //ROS_INFO("localizerEKF_2D::handleOdom dx,dy,dYaw,dt=[%lf,%lf,%lf,%lf]",dx,dy,dYaw,dt);
    double Vm=sqrt(dx*dx+dy*dy)/dt;
    double Wm=dYaw/dt;
    // Perform Propagate
    propagate(Vm,Wm,dt);
    // Publish the odometry with covariance topic
    geometry_msgs::Quaternion odom_quat = tf::createQuaternionMsgFromYaw(X(2));
    nav_msgs::Odometry outputPose;
    outputPose.header.frame_id = "odom";
    outputPose.header.stamp =odom->header.stamp;
    outputPose.pose.pose.position.x = X(0);
    outputPose.pose.pose.position.y = X(1);
    outputPose.pose.pose.position.z = 0;
    outputPose.pose.pose.orientation=odom_quat;
    for (unsigned int i=0; i<2; i++)
      for (unsigned int j=0; j<2; j++)
    	outputPose.pose.covariance[6*i+j] = P(i,j);
    posePub.publish(outputPose);
  }
  oldX=x;
  oldY=y;
  oldYaw=yaw;
  oldStamp=odom->header.stamp;
  // ROS_INFO("localizerEKF_2D::handleOdom ending");
}

void localizerEKF_2D::handleOdomSim(const nav_msgs::Odometry::ConstPtr& odom) {

  double x=odom->pose.pose.position.x;
  double y=odom->pose.pose.position.y;
  //  double yaw=acos(odom->pose.pose.orientation.w)*2.0;
  tf::Pose pose;
  tf::poseMsgToTF(odom->pose.pose, pose);
  double yaw=tf::getYaw(pose.getRotation());
  // ROS_INFO("localizerEKF_2D::handleOdom received [%lf,%lf,%lf]",x,y,yaw*180/M_PI);
  if(!initOdom) {
    oldX=x;
    oldY=y;
    oldYaw=yaw;
    initOdom=true;
    oldStamp=odom->header.stamp;
    X<<x,y,yaw;
    //  ROS_INFO("localizerEKF_2D::handleOdom X initialized to [%lf,%lf,%lf]",X(0),X(1),X(2)*180/M_PI);
    return;
  }
  if(filterOn) {
    if((oldX==x)&&(oldY==y)&&(oldYaw==yaw)){
      oldStamp=odom->header.stamp;
      //      ROS_INFO("localizerEKF_2D::handleOdom same position ending");
      return;
    }
    // Calculate the measurement
    double dx=x-oldX;
    double dy=y-oldY;
    double dYaw=angleDiff(yaw,oldYaw);
    double dt=odom->header.stamp.toSec()-oldStamp.toSec();
    //ROS_INFO("localizerEKF_2D::handleOdom dx,dy,dYaw,dt=[%lf,%lf,%lf,%lf]",dx,dy,dYaw,dt);
    double Vm=sqrt(dx*dx+dy*dy)/dt;
    double Wm=dYaw/dt;
    // Perform Propagate
    propagate(Vm,Wm,dt);
    // Publish the odometry with covariance topic
    geometry_msgs::Quaternion odom_quat = tf::createQuaternionMsgFromYaw(X(2));
    nav_msgs::Odometry outputPose;
    outputPose.header.frame_id = "odom";
    outputPose.header.stamp =odom->header.stamp;
    outputPose.pose.pose.position.x = X(0);
    outputPose.pose.pose.position.y = X(1);
    outputPose.pose.pose.position.z = 0;
    outputPose.pose.pose.orientation=odom_quat;
    for (unsigned int i=0; i<2; i++)
      for (unsigned int j=0; j<2; j++)
    	outputPose.pose.covariance[6*i+j] = P(i,j);
    posePub.publish(outputPose);
  }
  oldX=x;
  oldY=y;
  oldYaw=yaw;
  oldStamp=odom->header.stamp;
  // ROS_INFO("localizerEKF_2D::handleOdom ending");
}

void localizerEKF_2D::propagate(double Vm, double Wm, double dt) {
  //ROS_INFO("localizerEKF_2D::propagate start ");
  // Propagate the state beased on the motion model of linear velocity Vm and angular velocity Wm
  X(0)=X(0)+Vm*dt*cos(X(2));
  X(1)=X(1)+Vm*dt*sin(X(2));
  X(2)=thetapp(X(2)+Wm*dt);
  // Create the F matrix.
  Matrix3d Fr;
  Fr<<1, 0, -Vm*dt*sin(oldYaw),
    0, 1,  Vm*dt*cos(oldYaw),
    0, 0,  1;
  // Create the G matrix
  MatrixXd Gr(3,2);
  Gr<< -dt*cos(oldYaw), 0, 
    -dt*sin(oldYaw), 0,
    0,             -dt;
  //Propagate the covatiance of the robot
  P.block(0,0,3,3)=Fr*P.block(0,0,3,3)*Fr.transpose()+Gr*Qr*Gr.transpose();
  // Update the cumulative F function for batch propagation of the landmarks
  CF=Fr*CF;
  normalizeCovariance();
  //ROS_INFO("localizerEKF_2D::propagate end ");
}

bool localizerEKF_2D::processStart(localizer::StartFilter::Request  &req,
				localizer::StartFilter::Response &res) {
  if(req.start) {
    ROS_INFO("localizerEKF_2D::StartFilter received true");
  }
  else  {
    ROS_INFO("localizerEKF_2D::StartFilter received false");
  }
  filterOn=req.start;
  return true;
}

void localizerEKF_2D::propagateRL(){
  // Propagate the cross-corelation between robot and landmarks up to now
  // ROS_INFO("localizerEKF_2D::propagateRL start");
  for(int i=0; i<nL; i++) {
    P.block(0,3+i*2,3,2)=CF*P.block(0,3+i*2,3,2);
    P.block(3+i*2,0,2,3)=P.block(3+i*2,0,2,3)*CF.transpose();
  }
  // Reset the cumulative propagation F function.
  CF<<1,0,0,
    0,1,0,
    0,0,1;
  // ROS_INFO("localizerEKF_2D::propagateRL end");
}

bool localizerEKF_2D::processMeetpoint(localizer::UpdateFilter::Request  &req,
				    localizer::UpdateFilter::Response &res) {

  //ROS_INFO("localizerEKF_2D::processMeetpoint received meetpoint %d [%lf,%lf,%lf]", req.id, req.x, req.y);
  //ROS_INFO("localizerEKF_2D::processMeetpoint robot pose [%lf,%lf,%lf]",X(0),X(1),X(2));
  if(nL==0) {
    //If it is the first meetpoint, reset the odometry covariance.
    P=MatrixXd::Zero(3,3);
  }
  // Propagate the cross-corelation between robot and the 
  // existing landmarks up to now
  propagateRL();

  // For the new landmark
  ROS_INFO("localizerEKF_2D::processMeetpoint Prelimin");

  Vector2d xl; //the coordinates of the landmark (meetpoint) in World Frame
  Vector2d z; // Vector representation of the measurment
  z<<req.x, req.y;
  xl=X.segment(0,2)+C(X(2))*z;

  // Create the measurment matrices
  MatrixXd Hr(2,3);
  Vector2d dx;
  dx=xl-X.segment(0,2);
  double _c=cos(X(2));
  double _s=sin(X(2));
  Hr<<-_c, -_s, + _c*dx(1)-_s*dx(0),
    _s, -_c, - _s*dx(1)-_c*dx(0);
  Matrix2d Hl; // Hl is a 2 by 2 Matrix
  Hl=CT(X(2));
  //  ROS_INFO("localizerEKF_2D::processMeetpoint Check for id %d",req.id);
  if(newLandmark(req.id)) {
    // New Landmark
    //  ROS_INFO("localizerEKF_2D::processMeetpoint New Landmark");
    idList.push_back(req.id);
    //ROS_INFO("localizerEKF_2D::processMeetpoint  Expand");
    P.conservativeResize(P.rows()+2,P.cols()+2);
    X.conservativeResize(X.size()+2);
    X.segment(3+nL*2,2)=xl;
    // ROS_INFO("localizerEKF_2D::processMeetpoint Create new Pllcv");
    // Create new Pllcv matrices describing the cross corelation 
    // between the old landmarks and the new Landmark. Page 79 eq. 5.73 
    for (int count=0; count<nL; count++){
      //  ROS_INFO("localizerEKF_2D::processMeetpoint Pll for landmark %d %d",count, req.id);
      Matrix2d newPllcv;
      newPllcv=-P.block(3+count*2,0,2,3)*Hr.transpose()*Hl;
      P.block(3+count*2,3+nL*2,2,2)=newPllcv;
      P.block(3+nL*2,3+count*2,2,2)=newPllcv.transpose();    
    }
    // Create new Prl matrix describing the cross corelation 
    // between the Robot and the new Landmark. Page 79 eq. 5.73 
    //ROS_INFO("localizerEKF_2D::processMeetpoint Create new Prl matrix");
    MatrixXd newPrl(3,2);
  
    // cout<<-P.block(0,0,3,3)<<endl;
    // cout<<Hr.transpose()<<endl;
    // cout<<Hl<<endl;
    
    newPrl=-P.block(0,0,3,3)*Hr.transpose()*Hl;
    P.block(0,3+nL*2,3,2)=newPrl;
    P.block(3+nL*2,0,2,3)=newPrl.transpose();
    // Create the new Pll matrix describing the covariance of the landmark
    // page 79 eq. 5.74
    //ROS_INFO("localizerEKF_2D::processMeetpoint Create the new Pll matrix");
    Matrix2d newPll;
    newPll=Hl.transpose()*(Hr*P.block(0,0,3,3)*Hr.transpose()+R)*Hl;
    P.block(3+nL*2,3+nL*2,2,2)=newPll;
    nL++;
  }
  else {
    // Find which landmark it is:
    int id=-1;
    for(unsigned int i=0; i<idList.size(); i++){
      if(idList[i]==req.id)
	id=i;
    }
    //Sanity check:
    if(id<0) {
      cerr<<"Not able to find id:"<<req.id<< "in the list of ids"<<endl;
      for(unsigned int i=0; i<idList.size(); i++){
	cerr<<idList[i]<<" ";
      }
      cerr<<endl;
    }
    // perform update

    Vector2d z_est;
    z_est=CT(X(2))*(X.segment(3+2*id,2)-X.segment(0,2));

    Vector2d r;
    r=z-z_est;
    // cout<<"z     "<<z<<endl;
    // cout<<"Z_est "<<z_est<<endl;
    // cout<<"r     "<<r<<endl;
    MatrixXd H(2,3+2*nL);
    H=MatrixXd::Zero(2,3+2*nL);

    H.block(0,0,2,3)=Hr;
    H.block(0,3+2*id,2,2)=Hl;
    MatrixXd S(2,2);
    S=H*P*H.transpose()+R;
    MatrixXd K(2,3+2*nL);
    K=P*H.transpose()*S.inverse();
    X=X+K*r;
    X(2)=thetapp(X(2));
    P=(MatrixXd::Identity(3+nL*2,3+nL*2)-K*H)*P*(MatrixXd::Identity(3+nL*2,3+nL*2)-K*H).transpose()+K*R*K.transpose();
  }
  // publish the odometry with covariance topic    
  
  // Publish the Map
  //publishMap();
  localizer::GVGmap theMap;
  //theMap.frame_id = "map";
  //theMap.stamp =oldStamp;
  cout<<"+++++++++ "<<X.size()<<" ";
  for(unsigned int i=0; i<X.size(); i++) {
    theMap.state.push_back(X[i]);
    res.X.push_back(X[i]);
    cout<<X[i]<<" ";
  }
  cout<<endl;

  for(unsigned int i=0; i<P.rows(); i++) {
    for(unsigned int j=0; j<P.cols(); j++) {
      theMap.cov.push_back(P(i,j));
      res.P.push_back(P(i,j));
    }
  }
  mapPub.publish(theMap);
  // cerr<<"The id list:\n";
  // cerr<<"--------------------------------------------------------------\n";
  // for(unsigned int i=0; i<idList.size(); i++){
  //   cerr<<idList[i]<<" ";
  // }
  // cerr<<endl;
  //ROS_INFO("localizerEKF_2D::processMeetpoint end");
  return(true);
}
double localizerEKF_2D::thetapp(double theta) {
  while(theta>M_PI) theta-=2.0*M_PI;
  while(theta<-M_PI) theta+=2.0*M_PI;
  return(theta);
}
double localizerEKF_2D::theta02p(double theta) {
  theta=fmod(theta,2.0*M_PI);
  // while(theta>2*M_PI) theta-=2.0*M_PI;
  // while(theta<0) theta+=2.0*M_PI;
  return(theta);
}

bool localizerEKF_2D::newLandmark(int id){
  // ROS_INFO("localizerEKF_2D::newLandmark");
  for (vector<int>::iterator it = idList.begin(); it != idList.end(); ++it){
    if((*it)==id) return(false);
  }
  // ROS_INFO("localizerEKF_2D::newLandmark");
  return true;
}

// insideEllipse() {
// for ellipse with a and b semi axis oriented by angle t:
// ((x-xc)*cos(t)-(y-yc)*sin(t)).^2/a^2 + ...
// ((x-xc)*sin(t)+(y-yc)*cos(t)).^2/b^2 <= 1
//}

double localizerEKF_2D::angleDiff(double a, double b) {
  double diff2= theta02p(a)-theta02p(b);
  while (diff2>M_PI)  diff2-=2*M_PI; 
  while (diff2<-M_PI) diff2+=2*M_PI;
  return(diff2);
}

void localizerEKF_2D::normalizeCovariance(){
  for(int i=0; i<P.rows()-1; i++) {
    for (int j=i+1; j<P.cols(); j++){      
      double tmp=(P(i,j)+P(j,i))/2.0;
      P(i,j)=tmp;
      P(j,i)=tmp;
    }
  }
}

Matrix2d localizerEKF_2D::CT(double angle) {
  Matrix2d c;
  c<<cos(angle), sin(angle),
    -sin(angle), cos(angle);
  return(c);
}

Matrix2d localizerEKF_2D::C(double angle) {
  Matrix2d c;
  c<<cos(angle), -sin(angle),
    sin(angle), cos(angle);
  return(c);
}

bool localizerEKF_2D::processMaxUnc(localizer::MaxUncertainty::Request  &req,
				 localizer::MaxUncertainty::Response &res)  {
  double maxT=0;
  res.id=-1;
  for(int i=0; i<nL; i++) {
    double tmpT=P.block(3+i*2,3+i*2,2,2).trace();
    if(tmpT>maxT) {
      maxT=tmpT;
      res.id=i;
    }
  }
  return true;  
}

bool localizerEKF_2D::processMinUnc(localizer::MinUncertainty::Request  &req,
				 localizer::MinUncertainty::Response &res)  {
  double minT=1000000;
  res.id=-1;
  for(int i=0; i<nL; i++) {
    double tmpT=P.block(3+i*2,3+i*2,2,2).trace();
    if(tmpT<minT) {
      minT=tmpT;
      res.id=i;
    }
  }
  return true;
}
